/*
mqflash.c Version 0.2.0. ISO8859 Marquee in X Windows
for 2 line flashing display.
Copyright (C) 2001-2010  dondalah721@yahoo.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:

	Free Software Foundation, Inc.
	59 Temple Place - Suite 330
	Boston, MA  02111-1307, USA.
*/

/* to define the escape key */
#define XK_MISCELLANY 1

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/keysymdef.h>
#include <assert.h>
#include <unistd.h>
#include "12x24.h"

#define FONT "6x13"
#define NIL (0)
#define CIRCLE 23040
#define SZ 1

int whiteColor,blackColor;
int runflg,expflg;
unsigned long yellow;
Display *dpy;
Window w;
GC gc;
Font fontid;
Screen *dfltscr;
Colormap cmap;
XColor scrdef,exctdef;
int rootww;                      /* width  of root    window */
int rootwh;                      /* height of root    window */
int dpywdth;                     /* width  of display window */
int dpyhght;                     /* height of display window */
int charwdth;                    /* width of message character */

/* X Windows code is based on: */
/* http://tronche.lri.fr:8000/gui/x/xlib-tutorial/2nd-program-anatomy.html */

void getkey()
   {
   int msk;
   int symbol;
   int XCheckMaskEvent();
   XEvent e;
   XKeyEvent *k;

   msk = KeyPressMask|ExposureMask;

   XSelectInput(dpy, w, msk);

   while (XCheckMaskEvent(dpy, msk, &e))
      {
      if (e.type == KeyPress)
         {
         k = (XKeyEvent *) &e;
         symbol = XLookupKeysym(k,0);
         if (symbol == XK_Escape)
            {
            runflg = 0;
            } /* if quit */
         } /* if keypress event */
      else if (e.type == Expose)
         {
         expflg = 1;
         } /* if expose event */
      } /* if event received */
   } /* getkey */

void initx()
   {
   int rslt;
   char title[64];

   dpy = XOpenDisplay(NIL);

   if (dpy == NULL)
      {
      fprintf(stderr,"X Windows failure\n");
      exit(1);
      } /* if X Windows is not active */

   assert(dpy);

   rootww = XDisplayWidth(dpy,0);
   rootwh = XDisplayHeight(dpy,0);
   dpywdth = rootww -  80;
   dpyhght = rootwh - 100;

   whiteColor = WhitePixel(dpy, DefaultScreen(dpy));
   blackColor = BlackPixel(dpy, DefaultScreen(dpy));

   w = XCreateSimpleWindow(dpy,
      DefaultRootWindow(dpy),
      0, 0, 
      dpywdth, dpyhght,
      0, whiteColor,
      whiteColor);

   XSelectInput(dpy, w, StructureNotifyMask);

   XMapWindow(dpy, w);

   gc = XCreateGC(dpy, w, 0, NIL);

   fontid = (Font) XLoadFont(dpy,FONT);

   XSetFont(dpy,gc,fontid);

   XSetForeground(dpy, gc, blackColor);

   dfltscr = XDefaultScreenOfDisplay(dpy);
   if (dfltscr == NULL)
      {
      fprintf(stderr,"XDefaultScreenOfDisplay failed\n");
      perror("XDefaultScreenOfDisplay failed");
      exit(1);
      } /* if error */

   cmap = XDefaultColormapOfScreen(dfltscr);

   rslt = XAllocNamedColor(dpy,cmap,"yellow",
      &scrdef,&exctdef);
   if (rslt < 0)
      {
      fprintf(stderr,"XAllocNamedColor failed\n");
      perror("XAllocNamedColor failed");
      exit(1);
      } /* if error */
   yellow = scrdef.pixel;

   XSetWindowBorderWidth(dpy, w, 40);

   sprintf(title,"mqflash");
   XStoreName(dpy,w,title);

   XMoveWindow(dpy,w,0,0);

   while(1)
      {
      XEvent e;
      XNextEvent(dpy, &e);
      if (e.type == MapNotify) break;
      } /* wait for window initialization */

   XSetForeground(dpy, gc, blackColor);

   XDrawRectangle(dpy,w,gc,0,0,dpywdth,dpyhght);

   XFillRectangle(dpy,w,gc,0,0,dpywdth,dpyhght);

   } /* initx */

int getbyte()
   {
   int len;
   unsigned char buf[8];
   unsigned char *p;
   p = buf;
   len = read(0,buf,1);
   if (len == 0)
      return(EOF);
   else if (len < 0)
      {
      fprintf(stderr,"Error reading stdin\n");
      perror("Read error");
      exit(1);
      } /* if read err */
   else
      return(*p);
   } /* getbyte */

void putmq(unsigned char ch,int x,int y)
   {
   int i,j,col,row;
   char *p;
   p = (char *) font + (ch * 12 * 24);
   for (j=0;j<24;j++)
      {
      for (i=0;i<12;i++)
         {
         if (*p == 1)
            {
            col = x + (i * 4);
            row = y + (j * 4);
            XDrawArc(dpy,w,gc,col,row,SZ,SZ,0,CIRCLE);
            XFillArc(dpy,w,gc,col,row,SZ,SZ,0,CIRCLE);
            }
         p++;
         } /* for each col */
      } /* for each row */
   } /* putmq */

void putflash(unsigned char *line1,
   unsigned char *line2, unsigned char *line3)
   {
   int i,col,len;
   unsigned char *p,*q;
   col = 10;
   len = strlen((char *) line1);
   p = (unsigned char *) line1;
   for (i=0;i<len;i++)
      {
      q = (unsigned char *) p+i;
      putmq(*q,col,100);
      col += charwdth;
      }
   col = 10;
   len = strlen((char *) line2);
   p = (unsigned char *) line2;
   for (i=0;i<len;i++)
      {
      q = (unsigned char *) p+i;
      putmq(*q,col,235);
      col += charwdth;
      }
   col = 10;
   len = strlen((char *) line3);
   p = (unsigned char *) line3;
   for (i=0;i<len;i++)
      {
      q = (unsigned char *) p+i;
      putmq(*q,col,360);
      col += charwdth;
      }
   } /* putflash */

void putmsg(unsigned char *str)
   {
   int i,eomflg;
   int lmt;
   unsigned char *p,*q,*r,*lastspc;
   unsigned char line1[128],line2[128],line3[128];
   lmt = dpywdth / charwdth;
   p = (unsigned char *) str;        /* current pointer in message */
   if (*p == '\0')
      {
      fprintf(stderr,"Empty message\n");
      exit(1);
      } /* empty message */
   while (runflg)
      {
      XSetForeground(dpy, gc, blackColor);
      XDrawRectangle(dpy,w,gc,0,0,dpywdth,dpyhght);
      XFillRectangle(dpy,w,gc,0,0,dpywdth,dpyhght);
      XSetForeground(dpy, gc, yellow);
      eomflg = 0;
      line1[0] = line2[0] = line3[0] = '\0';
      /* first line */
      while (*p)
         {
         if (*p == ' ') p++;
         else break;
         } /* skip leading blanks */
      if (*p == '\0')
         {
         p = str;     /* repeat message on marquee */
         continue;
         } /* if end of message */
      lastspc = NULL;
      r = (unsigned char *) line1;
      for (i=0;i<lmt;i++)
         {
         q = (unsigned char *) p+i;
         if (*q == ' ')
	    lastspc = (unsigned char *) r;
         else if (*q == '\0')
            {
            p = (unsigned char *) str;
            eomflg = 1;
            break;
            } /* if end of message */
         *r++ = *q;
         } /* format line 1 */
      if (eomflg)
         {
	 putflash(line1,line2,line3);
	 p = (unsigned char *) str;
	 continue;
	 } /* if eom on line1 */
      if (lastspc != NULL) *lastspc = '\0';
      else *r = '\0';
      while (*q != ' ' && q > p) q--;
      p = (unsigned char *) q;

      /* second line */
      while (*p)
         {
         if (*p == ' ') p++;
         else break;
         } /* skip leading blanks */
      lastspc = NULL;
      r = (unsigned char *) line2;
      for (i=0;i<lmt;i++)
         {
         q = (unsigned char *) p+i;
         if (*q == ' ') lastspc = r;
         else if (*q == '\0')
            {
            p = (unsigned char *) str;
	    eomflg = 1;
            break;
            } /* if end of message */
         *r++ = *q;
         } /* format line 2 */
      if (eomflg)
         {
	 putflash(line1,line2,line3);
	 p = (unsigned char *) str;
	 continue;
	 } /* if eom on line2 */
      if (lastspc != NULL) *lastspc = '\0';
      else *r = '\0';
      while (*q != ' ' && q > p) q--;
      p = (unsigned char *) q;

      /* third line */
      while (*p)
         {
         if (*p == ' ') p++;
         else break;
         } /* skip leading blanks */
      lastspc = NULL;
      r = line3;
      for (i=0;i<lmt;i++)
         {
         q = (unsigned char *) p+i;
         if (*q == ' ') lastspc = r;
         else if (*q == '\0')
            {
            p = (unsigned char *) str;
	    eomflg = 1;
            break;
            } /* if end of message */
         *r++ = *q;
         } /* format line 3 */
      if (lastspc != NULL) *lastspc = '\0';
      else *r = '\0';
      while (*q != ' ' && q > p) q--;
      p = (unsigned char *) q;

      putflash(line1,line2,line3);
      for (i=0;i<6;i++)
         {
         getkey();
         if (!runflg) break;
         if (expflg)
            {
            expflg = 0;
            XSetForeground(dpy, gc, blackColor);
            XDrawRectangle(dpy,w,gc,0,0,dpywdth,dpyhght);
            XFillRectangle(dpy,w,gc,0,0,dpywdth,dpyhght);
            XSetForeground(dpy, gc, yellow);
            } /* if exposed */
         sleep(1);
         } /* wait loop */
      } /* msg loop */
   } /* putmsg */

int main()
   {
   int ch;
   int i;
   unsigned char *str;
   unsigned char *p,*q,*r;
   str = (unsigned char *) malloc(1024*1024);
   if (str == NULL)
      {
      fprintf(stderr,"Not enough memory.\n");
      exit(1);
      } /* not enough memory */
   initx();
   runflg = 1;
   expflg = 0;
   charwdth = 50;
   p = (unsigned char *) str;
   q = (unsigned char *) str + 12;
   while (p < q) *p++ = ' ';
   p = (unsigned char *) str+12;
   q = (unsigned char *) str + (1024*1024) - 64;
   while (p < q)
      {
      ch = getbyte();
      if (ch == EOF) break;
      else if (ch < ' ') ch = ' ';
      else if (ch == 127) ch = ' ';
      *p++ = (unsigned char) ch;
      } /* fill str */
   *p = '\0';
   r = (unsigned char *) str;
   i = p - r;
   if (i == 12) return(0);
   q = (unsigned char *) p + 12;
   while (p < q) *p++ = ' ';
   *p = '\0';
   putmsg(str);
   return(0);
   } /* main */
